"""
Copyright (c) 2009 [name + "(at)student.hpi.uni-potsdam.de" for name in ["murat.knecht", "franziska.haeger", "christian.schwarz"]]

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pygame, time, astar

from enemy import Enemy
from level import Level
from vector2D import Vector2D
from tile import Tile
from resourcemanager import ResourceManager
from path import Pathfinder
from animations import ResourceAnimation
import events, cfg

from chat import ChatBox

from controller import Controller
class EnemyController(Controller):

	def __init__(self, gamecontroller, level, endlessmode=False):
		Controller.__init__(self)
		self.gamecontroller = gamecontroller
		self.level = level
		self.surface = pygame.Surface(self.level.pxl_size, cfg.SURFACEFLAGS)
		
		# used for endlessmode
		self.endlessmode = endlessmode
		self.wavecircle = 0
		self.updateHP = 0
		
		#pathfidner for the enemies
		self.pathfinder = Pathfinder()
		self.pathfinder.set_level(level)
		self.pathfinder.start()
		
		self.waves = level.nbrOfWaves
		self.activeWave = 0
		self.cooldown = self.level.waves[self.activeWave]["timeout"]
		self.enemyGroup = pygame.sprite.Group()
	
	@property
	def image(self):
		return self.surface

	def initialize_nextLevel(self, level):
		self.level = level
		self.pathfinder.set_level(level)
		
		self.waves = level.nbrOfWaves
		self.activeWave = 0
		self.cooldown = self.level.waves[self.activeWave]["timeout"]
		self.enemyGroup = pygame.sprite.Group()
	
	def get_wave_enemies(self, wave):
		enemies = []
		
		for idx in xrange(wave["count"]):
			hp = wave["hp"] + self.updateHP
			enemy = Enemy(self, wave["enemy"], self.activeWave, hp, wave["speed"], wave["res"])
			enemies.append(enemy)
		
		return enemies, hp
	
	def nearest_enemy(self, pos, attackables):
		"""Returns the nearest enemy to the given position."""
		if not len(self.enemyGroup):
			return None
		
		enemies = filter(lambda enemy: enemy.movement in attackables, self.enemyGroup)
		if not enemies:
			return None
		
		return min(enemies, key = lambda enemy: Vector2D(enemy.rect.center).sub(pos).magnitude())
	
	def hurt_enemy(self, enemy, damage, damagetype):
		"""Inflicts the given damage to the enemy, considering resistance.
		
		@return Returns the amount of damage actually inflicted.
		"""
		# Reduce the damage by the bug's resistance against the damage's type.
		resistance = int(damage * enemy.get_resistance_to(damagetype) / 100)
		inflicted_damage = max(0, damage - resistance)
		
		# Inflict the damage to the bug's health points.
		enemy.hp = max(0, enemy.hp - inflicted_damage)
		
		# If it is not dead, we're done.
		if enemy.hp > 0:
			return inflicted_damage
		
		# If it died, we need to handle its death.
		self._kill(enemy)
		
		return inflicted_damage

	@property
	def noMoreEnemies(self):
		return len(self.enemyGroup) == 0 and self.noAdditionalWave
	
	def enemies_in_range(self, func):
		"""Provide for the given tile coordinates all enemies standing on it.
		
		@param func the filter function
		@return a possibly empty list of enemies.
		"""		
		return filter(func, self.enemyGroup)
	
	@property
	def noAdditionalWave(self):
		return self.activeWave >= self.level.nbrOfWaves and not self.endlessmode
	
	def update(self, tick):
		self.cooldown -= tick
		
		enemyHP = 0
		if self.cooldown <= 0 and not self.noAdditionalWave:
			
			# new Wave!
			wave = self.level.waves[self.activeWave]
			enemies, enemyHP = self.get_wave_enemies(wave)
			if self.activeWave+1 < self.level.nbrOfWaves:
				self.cooldown = self.level.waves[self.activeWave+1]["timeout"]
			
			if wave.has_key("way"):
				way = wave["way"]
			else:
				way = self.level.endpoint
				
			# Post the wave's message, if any.
			ChatBox.post_system_message(enemies[0].displayname + "s incoming!")
			if wave["message"]:
				ChatBox.post_system_message(wave["message"])
			
			for creep in enemies:
				start = self.level.startTile
				creep.set_position(self.level.tile_to_pxl(Vector2D(start)))
				creep.nextPathNode = 0
				
				creep.way = (way, start)
				if creep.movement == "air":
					creep.path = self.pathfinder.get_airpath(creep.way)
				else:
					creep.path = self.pathfinder.get_path(creep.way)
				creep.pathNode = creep.position
				self.enemyGroup.add(creep)
			self.activeWave += 1
		
		if self.endlessmode:
			if self.activeWave >= self.level.nbrOfWaves:
				self.activeWave = 0
				self.wavecircle += 1
				self.updateHP = enemyHP
				self.cooldown = self.level.waves[self.activeWave]["timeout"]
		
		if not self.endlessmode:
			self.gamecontroller.update_wave(str(self.activeWave) + " / " + str(self.level.nbrOfWaves))
		else:
			self.gamecontroller.update_wave(str(1 + self.activeWave + self.wavecircle * self.level.nbrOfWaves))
		
		self.gamecontroller.update_waveCooldown(self.cooldown)
		self.enemyGroup.update(tick)
		
		# Find out, if some bug reached the end
		for enemy in self.enemyGroup:
			if enemy.hasarrived():
				self.enemyGroup.remove(enemy)
				self.gamecontroller.update_lifes(-1)		
		
		# self.gamecontroller.update_lifes(-pygame.sprite.collide(self.targetsprite, enemySpriteGroup, True, pygame.sprite.collide_rect))

	def handle_mouse_hovering(self, pos):
		if pos:
			pos = self.recalculate_position(pos)
			for enemy in self.enemyGroup:
				enemy.hover = enemy.rect.collidepoint(pos)
		else:
			for enemy in self.enemyGroup:
				enemy.hover = False

	def draw(self):
		self.surface.fill((0,0,0,0))
		self.enemyGroup.draw(self.surface)
		
		for enemy in self.enemyGroup:
		# Show range and Update Cost
			if enemy.hover:
				
				pygame.draw.rect(self.surface, (255,0,0), pygame.Rect(enemy.rect.centerx, enemy.rect.centery-10, 20, 5))
				if enemy.hp > 0:
					w = enemy.hp * 20 / enemy.maxHp
				pygame.draw.rect(self.surface, (0,255,0), pygame.Rect(enemy.rect.centerx, enemy.rect.centery-10, w, 5))

	def get_enemies_on_tiles(self, tiles):
		"""Returns all enemies that are situated on the tiles.
		
		@param tiles Iterable of tiles
		@return list of enemies"""
		return [ enemy for enemy in self.enemyGroup if self.level.pxl_to_tile_coordinates(Vector2D(enemy.position)) in tiles ]


	def kill_em_all(self):
		"""Cheat: Kill all enemies on the map and collect the money."""
		for enemy in self.enemyGroup:
			self._kill(enemy)
			
	def next_wave_plz(self):
		"""Cheat: Start the next wave now."""
		self.cooldown = 1
	
	def _kill(self, enemy):
		self.enemyGroup.remove(enemy)
		self.gamecontroller.update_score(enemy.waveNr + 1 + self.wavecircle * self.level.nbrOfWaves)
		events.add_resources(enemy.res)
		events.add_animation(ResourceAnimation(enemy.rect.center, enemy.res))
		enemy.death_sound.play()
	