"""
############################################################################
# Copyright (c) 2009 T-002@gmx.de                                          #
#                                                                          #
# Permission is hereby granted, free of charge, to any person              #
# obtaining a copy of this software and associated documentation           #
# files (the "Software"), to deal in the Software without                  #
# restriction, including without limitation the rights to use,             #
# copy, modify, merge, publish, distribute, sublicense, and/or             #
# sell copies of the Software, and to permit persons to whom the           #
# Software is furnished to do so, subject to the following conditions:     #
#                                                                          #
# The above copyright notice and this permission notice                    #
# shall be included in all copies or substantial portions of the Software. #
#                                                                          #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,          #
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF       #
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.   #
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY     #
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,     #
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE        #
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                   #
############################################################################
"""
__license__ = __doc__

import math
from vector2D import Vector2D

class Gesture(object):
	"""The main class of PyGestures.
	
	This class represents a gesture which can be used for gesture detection.
	
	@author T-002@gmx.de
	@version 1.4
	"""
	
	def __init__(self, itemList=None, tolerance=0.1):
		"""Initializes the gesture.
		
		@param tolerance Tolerance value used for gesture equality detection.
		@param itemList List of Vector2D or tuple instances which represent
		the gesture.
		"""
		object.__init__(self)
		
		self.tolerance = tolerance
		self.finalized = False
		
		self.points = []
		
		if itemList:
			for item in itemList:
				self.add_point(item)
	
	@property
	def maxX(self):
		"""Returns the heighest x value of the gesture.
		
		@return X coordinates of the heighest x value. If the points list
		is empty 0 will be returned.
		"""
		if self.points:
			return max(self.points, key = lambda point: point.x).x
		else:
			return 0
	
	@property
	def minX(self):
		"""Returns the lowest x value of the gesture.
		
		@return X coordinates of the lowest x value. If the points list
		is empty 0 will be returned.
		"""
		if self.points:
			return min(self.points, key = lambda point: point.x).x
		else:
			return 0
	
	@property
	def maxY(self):
		"""Returns the heighest y value of the gesture.
		
		@return Y coordinates of the heighest y value. If the points list
		is empty 0 will be returned.
		"""
		if self.points:
			return max(self.points, key = lambda point: point.y).y
		else:
			return 0
	
	@property
	def minY(self):
		"""Returns the lowest y value of the gesture.
		
		@return Y coordinates of the lowest y value. If the points list
		is empty 0 will be returned.
		"""
		if self.points:
			return min(self.points, key = lambda point: point.y).y
		else:
			return 0
	
	@property
	def center(self):
		"""Returns a Vector2D instance representing the center of the gesture.
		
		This function should be called before the finalize() call.
		
		@return Returns a Vector2D representing the center of the gesture."""
		centerX = ((self.maxX - self.minX) / 2.0) + self.minX
		centerY = ((self.maxY - self.minY) / 2.0) + self.minY
		
		return Vector2D(centerX, centerY)
	
	@property
	def startPoint(self):
		"""Returns a Vector2D instance representing the startpoint of the gesture.
		
		This function should be called before the finalize() call.
		
		@return Vector2D instance representing the current startpoint of the gesture.
		"""
		if self.points:
			return self.points[0]
		else:
			return None
	
	@property
	def endPoint(self):
		"""Returns a Vector2D instance representing the last point of the gesture.
		
		This function should be called before the finalize() call.
		
		@return Vector2D instance representing the current endpoint of the gesture.
		"""
		if self.points:
			return self.points[-1]
		else:
			return None
	
	@property
	def width(self):
		"""Returns the width of the gesture.
		
		@return Returns the width of the gesture.
		"""
		return (self.maxX - self.minX)

	@property
	def height(self):
		"""Returns the heigth of the gesture.
		
		@return Returns the heigth of the gesture.
		"""
		return (self.maxY - self.minY)
	
	def add_point(self, item):
		"""Adds a point to the gesture.
		
		@param item Vector2D or tuple instance representing the point.
		If the item is the same then the last added, it would be ignored.
		"""
		self.finalized = False
		
		if not isinstance(item, Vector2D):
			item = Vector2D(item)
		
		if len(self.points) == 0:
			self.points.append(item)
		elif self.points[-1] != item:
			self.points.append(item)
	
	def calculate_distance(self, vec2D_1, vec2D_2):
		"""Returns the distance between the given points.
		
		@param vec2D_1, vec2D_2 Vector2D instances for distance calculation.
		
		@return Returns the distance between the given vectors.
		"""
		return abs(vec2D_1 - vec2D_2)
	
	def magnitude(self, pointlist=None):
		"""Finds the length of the stroke.
		
		@param pointlist If a point list is given, finds
						 the length of that list.
		
		@return Returns the length of the gesture, or the length of
				the given pointlist.
		"""
		if not pointlist:
			points = self.points
		
		length = 0.0
		
		if len(points) > 1:
			for idx in xrange(len(points) - 1):
				length += self.calculate_distance(points[idx], points[idx + 1])
		
		return length
	
	def set_sample_points(self, nbrOfPoints=32):
		"""Sets the number of samplepoints of the stroke.
		
		The points are placed in equal distance on the original
		gestures strokes.
		
		@param nbrOfPoints Number of points the gesture should have.
		
		@return True if stroke is normalized, False otherwise.
		"""        
		if len(self.points) <= 1 or self.magnitude() == 0.0:
			return False
		
		self.finalized = False
		
		targetLength = self.magnitude() / (nbrOfPoints - 1)
		result = [self.points[0]]
		
		idx = 0
		pointCount = len(self.points)
		
		while idx < pointCount:
			toTravel = targetLength
			startPoint = result[-1]
			
			for idx2 in range(idx, pointCount):
				if startPoint == self.points[idx2]:
					continue
				
				idx = idx2
				
				distance = (self.points[idx2] - startPoint).magnitude()
				toTravel -= distance
				
				if toTravel <= 0:
					break
				
				startPoint = self.points[idx2]
			
			if toTravel < 0:
				direction = (self.points[idx] - startPoint).normalized()
				toTravel += distance
				result.append(startPoint + (direction * toTravel))
			#perfect fit or last point
			elif toTravel == 0:
				result.append(self.points[idx])
			# last point
			elif toTravel > 0:
				# this is necessary, because of floating point precision
				if len(result) == nbrOfPoints:
					result[-1] = self.points[-1]
				else:
					result.append(self.points[idx])
				
				break
		
		self.points = result
		return True            
	
	def centralize(self):
		"""Sets the gestures center to (0, 0)"""
		self.points = self.offset(-self.center)
	
	def refinalize(self, tolerance):
		"""Refinalizes a gesture, read from the database.
		
		There is no check, if the gesture was already finalized!!!
		"""
		self.tolerance = tolerance
		self.finalized = True
		self.gestureProduct = self.dot_product(self)

	def finalize(self, nbrOfPoints=32):
		"""Finalizes the gesture for comparison.
		
		@param nbrOfPoints Number of points the gesture should have.
		
		@sideeffect If this function is called, the values of
					minX, minY, maxX, maxY, center, startPoint and endPoint
					are changed.					
		"""
		if self.finalized and len(self.points) == nbrOfPoints:
			return
		
		self.finalized = False
		
		self.set_sample_points(nbrOfPoints)
		self.centralize()
		
		size = max([abs(self.maxX -self.minX), abs(self.maxY - self.minY)])
		if size:
			self.points = self.scale(1.0 / size)
		
		self.gestureProduct = self.dot_product(self)
		
		self.finalized = True
	
	def dot_product(self, comparisonGesture):
		"""Returns the dot product between the gesture and the comparisonGesture.
		
		The dotproduct of a gesture is the sum of the dotproducts of the ordered
		points of the gestures.
		
		@param comparisonGesture Gesture used to calculate the dot product.
		
		@return Returns the dot product of the gestures.
		"""
		dotProduct = 0.0
		
		if len(self.points) != len(comparisonGesture.points):
			return -1
		
		for selfPoint, cmpPoint in zip(self.points, comparisonGesture.points):
			dotProduct += selfPoint * cmpPoint
		
		return dotProduct
	
	def get_score(self, comparisonGesture):
		""" Returns the matching score of the gesture against another gesture
		
		@return Returns an floatvalue representing the score.
		"""       
		if isinstance(comparisonGesture, Gesture):
			score = self.dot_product(comparisonGesture)
		
		if score < 0:
			return score
		
		score /= math.sqrt(self.gestureProduct * comparisonGesture.gestureProduct)
		
		return score
	
	def to_tuple_list(self):
		"""Returns a list of tuples representing the ordered list of points representing the gesture.
		
		@return Returns a list of tuples representing the ordered list of points representing the gesture.
		"""
		return [point.to_tuple() for point in self.points]
	
	def __eq__(self, comparisonGesture):
		"""Allows easy comparisons between gesture instances.
		
		If one of the gestures is not finalized, it will be finalized before
		comparison.
		
		@param comparisonGesture The gesture used for comparison.
		
		@return Returns True if both gestures match, False otherwise.
		
		@throw Raises a TypeError, if the comparisonGesture is not an instance
			   of Gesture.
		"""
		if isinstance(comparisonGesture, Gesture):
			if not self.finalized:
				self.finalize()
			if not comparisonGesture.finalized:
				comparisonGesture.finalize(len(self.points))
			
			score = self.get_score(comparisonGesture)
			return score > (1.0 - self.tolerance) and score < (1.0 + self.tolerance)
		else:
			raise TypeError("comparisonGesture has to be an instance of Gesture.")
	
	def __ne__(self, comparison_gesture):
		"""Allows easy comparisons between gesture instances.
		
		@param comparisonGesture The gesture used for comparison. It will
								 be finalized, if it was not done.
		
		@return Returns True if both gestures does not match, False otherwise.
		
		@throw Raises a TypeError, if the comparisonGesture is not an instance
			   of Gesture.
		"""
		return not self.__eq__(comparison_gesture)
	
	def scale(self, factor):
		"""Scales the gesture by the given factor.
		
		@param factor Factor used for scaling.
		
		@return Returns a list of points representing the scaled points
		of the gesture.
		"""
		return [point * factor for point in self.points]
	
	def __mul__(self, factor):
		"""Scales the gesture by the given factor.
		
		@param factor Factor used for scaling.
		
		@return Returns a new Gesture instance representing the scaled
		gesture.
		"""
		return Gesture(self.scale(factor), self.tolerance)
	
	def __div__(self, factor):
		"""Scales the gesture by the given factor.
		
		@param factor Factor used for scaling.
		
		@return Returns a new Gesture instance representing the scaled
		gesture.
		"""
		return self.__mul__(1.0 / factor)
	
	def offset(self, vec2D):
		"""Offsets the gesture by the given vec2D.
		
		@param vec2D Vector2D instance used to offset the gesture.
		
		@return Returns a list of offsetted points.
		"""
		return [point + vec2D for point in self.points]
	
	def __add__(self, other):
		"""Adds the item to the gestures.
		
		@param other If other is either a tuple or a Vector2D instance, every
					 point of the gets the offset. If other is another gesture
					 each element of the finalized self.points gets added with
					 the finalized other.points.
		
		@return Returns new gesture representing the result.
		
		@throw Can raise an TypeError, if other is neither a gesture, tuple
			   nor Vector2D instance.
		"""
		if isinstance(other, tuple):
			other = Vector2D(other)
		
		if isinstance(other, Vector2D):
			return Gesture(self.offset(other), self.tolerance)
		elif isinstance(other, Gesture):
			if not self.finalized:
				self.finalize()
			
			if not other.finalized:
				other.finalize(len(self.points))
			
			res = [selfpoint + otherpoint for selfpoint, otherpoint in zip(self.points, other.points)]
			
			return Gesture(res, self.tolerance)
		else:
			raise TypeError("other has to be a tuple, Vector2D instance or another gesture.")
	
	def __len__(self):
		"""Returns the length of the gesture.
		
		@return Returns the length of the gesture. The length is represented by the
				number of points the gesture contains.
		"""
		return len(self.points)