"""
Copyright (c) 2009 [name + "(at)student.hpi.uni-potsdam.de" for name in ["murat.knecht", "franziska.haeger", "christian.schwarz"]]

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from vector2D import Vector2D
from gesture import Gesture
import sqlite3.dbapi2 as db
import os

from chat import ChatBox

class GestureDetector(object):
	"""This class is used for gesture detection."""
	
	def __init__(self):
		"""Initiates the GestureDetector."""
		database = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "data", "gestures", "gesture.db")
		self.connection = self.get_database_connection(database)
		self.detaillevel = self.get_detaillevel_from_database()
		self.gestures = self.read_gestures_from_database()

	def execute_sql(self, statement, returnResult=False):
		"""Executes the given SQL statement.
		
		The database connection has to be stored in self.connection.
		
		@param statement Can be a single SQL statement or a list of statements
						 that should be executed in order.
		@param returnResult If this is True the result of the SQL statement is
		                    returned.
		
		@return If returnResult is True, the result of the last statement is
		        returned, None otherwise.
		"""
		cursor = self.connection.cursor()
		
		if isinstance(statement, basestring):
			cursor.execute(statement)
		elif isinstance(statement, list):
			for sql in statement:
				cursor.execute(sql)
		
		self.connection.commit()
		
		result = None
		if returnResult:
			result = cursor.fetchall()
		
		cursor.close()
		return result

	def get_database_connection(self, database):
		"""Opens a connection to the gesture database.
		
		@param database Database connection string.
		
		@return returns a connection to the database.
		"""
		if not os.path.exists(database):
			raise SystemExit("Cannot find gesture database: " + str(database))
		
		return db.connect(database)
	
	def get_detaillevel_from_database(self):
		"""Reads the detaillevel from the database.
		
		The database connection has to be stored in self.connection.
		
		@return Returns the detaillevel for detection.
		"""
		sql = "SELECT detaillevel FROM metainfo;"
		return self.execute_sql(sql, True)[0][0]
	
	def read_gestures_from_database(self):
		"""Reads the gestures from the database.
		
		The database connection has to be stored in self.connection.
		self.detaillevel has to be set!
		
		@return Returns a list containing the finalized gestures from the database.
		        The index of each gestures represents their ids in the database.
		"""
		sql = "SELECT * FROM gestures ORDER BY id;"
		dbgestures = self.execute_sql(sql, True)
		
		res = []
		for dbgesture in dbgestures:
			# dbgesture contains 2 * deteillevel + 2 elements (id and tolerance)
			assert(len(dbgesture) == 2 * self.detaillevel + 2)
			
			tolerance = dbgesture[1]
			
			gesture = Gesture(tolerance=tolerance)
			for idx in xrange(2, 2*self.detaillevel + 2, 2):
				gesture.add_point((dbgesture[idx], dbgesture[idx+1]))
			
			gesture.refinalize(tolerance+0.05)
			res.append(gesture)
		
		return res
	
	def compare_with_gestures(self, item):
		"""Compares the item with the gestures from the database.
		
		@param item Can either be a list of Vector2D instances, tuples or a gesture.
		
		@return Returns a dictionary containing the id, startPoint, endPoint, center,
		        width and height of the found gesture. If the item does not
		        match any gesture None will be returned.
		"""
		if not (isinstance(item, list) or isinstance(item, Gesture)):
			return None
		
		if isinstance(item, list):
			testGesture = Gesture(item)
		elif isinstance(item, Gesture):
			testGesture = item
		
		result = {"center": testGesture.center,
				  "width":  testGesture.width,
				  "height": testGesture.height,
				  "startPoint": testGesture.startPoint,
				  "endPoint": testGesture.endPoint
				  }
		
		resId = None
		for idx in xrange(len(self.gestures)):
			if self.gestures[idx] == testGesture:
				resId = idx
				break
		
		if resId == None:
			ChatBox.post_system_message("[GestureDetector] Gesture NOT found.")
			return None
		ChatBox.post_system_message("[GestureDetector] Gesture found.")
		
		result["id"] = resId
		
		return result