"""
Copyright (c) 2009 [name + "(at)student.hpi.uni-potsdam.de" for name in ["murat.knecht", "franziska.haeger", "christian.schwarz"]]

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pygame, os, random

import cfg

from configreader import ConfigReader
from resourcemanager import ResourceManager
from tile import TileFactory, Tile 
from vector2D import Vector2D

class Level(object):	
	def __init__(self, levelfile):
		"""Initiates the level.
		
		@param levelfile name of the lvl file containing the level definition.
		       Level files are stored  in data/maps.
		"""
		object.__init__(self)
		
		self.levelfile = ResourceManager.get_level_path(levelfile)
		self.mapTiles = []
		self.startpoints = []
		self.moveables = []
		self.wayTiles = []
		
		self.buildables = []
		self.nonbuildables = []
		
		self.endpoint = None
		self.waves = []
		self.bgimage = None
		self.image = None
		self.intro_image = None
		self.tiles = pygame.sprite.Group()
		
		self.parser = ConfigReader(self.levelfile)
		
		self.read_level_globals()
		self.parse_floorplan()
		self.create_images()
		self.prepare_waves()
	
	@property
	def pxl_width(self):
		return self.image.get_width()
	
	@property
	def pxl_height(self):
		return self.image.get_height()
	
	@property
	def pxl_size(self):
		return (self.pxl_width, self.pxl_height)
	
	@property
	def startTile(self):
		return random.choice(self.startpoints)
	
	@property
	def randomTile(self):
		return random.choice(self.moveables)
		
	def is_valid_tile_position(self, pos):
		"""Determine if the tile position is valid.
		
		If this method returns True, get_tile must return a valid tile.
		
		@param tuple or list with two elements
		@return True, in case the position is valid, False otherwise"""
		if not 0 <= pos[0] < self.size.x:
			return False
		
		if not 0 <= pos[1] < self.size.y:
			return False
		
		return True
		
	def get_tile(self, pos, y= None):
		if y == None:
			y = pos[1]
			pos = pos[0]
		return self.floorplan[pos][y]
	
	def tile_to_pxl(self, pos):
		res = Vector2D()
		res.x = pos[0] * Tile.vecsize[0]
		res.y = pos[1] * Tile.vecsize[1]
		return res
	
	def pxl_to_tile(self, pos):
		pos[0] = pos[0] / Tile.vecsize[0]
		pos[1] = pos[1] / Tile.vecsize[1]
		return pos
	
	def pxl_to_tile_coordinates(self, pos):
		"""Returns the tile coordinates, works in-place."""
		pos[0] = int(pos[0] / Tile.vecsize[0])
		pos[1] = int(pos[1] / Tile.vecsize[1])
		return pos
	
	def read_level_globals(self):
		"""Reads the level globals."""
		self.lifes      = self.parser.get_parameter("GLOBALS", "lifes")
		self.size       = Vector2D(self.parser.get_parameter("GLOBALS", "size"))
		self.nbrOfWaves = self.parser.get_parameter("GLOBALS", "waves")
		self.players    = self.parser.get_parameter("GLOBALS", "players")
		self.name       = self.parser.get_parameter("GLOBALS", "name")
		self.startresources = self.parser.get_parameter("GLOBALS", "startresources")
		
		if "image" in self.parser.get_parameter_names("GLOBALS"):
			self.image = ResourceManager.get_image(self.parser.get_parameter("GLOBALS", "image"))
		
		if "bgimage" in self.parser.get_parameter_names("GLOBALS"):
			self.bgimage = ResourceManager.get_image(self.parser.get_parameter("GLOBALS", "bgimage"))

		try:
			self.intro_image = ResourceManager.get_image(self.parser.get_parameter("GLOBALS", "intro"))
		except Exception,e:
			#print "no intro: " + str(e)
			self.intro_image = None

	def parse_floorplan(self):
		self.floorplan = []
		width = int(self.size.x)
		
		for line in xrange(int(self.size.y)):			
			counter = 0
			self.floorplan.append([])
			data = self.parser.get_parameter("GLOBALS", "floorplan_l" + str(line))
			for idx in xrange(len(data)):
				tile = TileFactory.get_tile(data[idx])
				
				tile.rect.topleft = (Tile.vecsize.x * counter, Tile.vecsize.y * line)
				
				self.floorplan[line].append(tile)
				self.tiles.add(tile)
				self.mapTiles.append((counter, line))
				if tile.target:
					self.endpoint = (counter, line)
				elif tile.lair:
					self.startpoints.append((counter, line))
				elif tile.movable: 
					self.moveables.append((counter, line))
				
				if tile.buildable:
					self.buildables.append(tile)
				else:
					self.nonbuildables.append(tile)
				
				counter += 1
			
			if len(self.floorplan[-1]) != int(self.size.x):
				raise SyntaxWarning("Level definition does not match the level size.\n%s instead of %s." % (len(self.floorplan[-1]), self.size.x))
		
		if len(self.floorplan) != int(self.size.y):
			raise SyntaxWarning("Level definition does not match the level size.\n%s instead of %s." % (len(line), self.size.x))

	def prepare_waves(self):
		"""Prepares the waves of the level."""
		self.waves = []
		
		for counter in xrange(1, self.nbrOfWaves+1):
			section = "WAVE_%s" % counter
			
			wave = {}
			wave["enemy"] = self.parser.get_parameter(section, "enemy")
			wave["count"] = self.parser.get_parameter(section, "count")
			wave["hp"] = self.parser.get_parameter(section, "hp")
			wave["res"] = self.parser.get_parameter(section, "res")
			wave["timeout"] = self.parser.get_parameter(section, "timeout")
			if self.parser._cp.has_option(section, "way"):
				wave["way"] = self.parser.get_parameter(section, "way")
				self.wayTiles.append(wave["way"])
			wave["speed"] = self.parser.get_parameter(section, "speed")
			self.waves.append(wave)
			if self.parser._cp.has_option(section, "message"):
				wave["message"] = self.parser.get_parameter(section, "message")
			else: wave["message"] = None
	
	def create_images(self):
		"""Creates the map images.
		
		@return Returns self.
		"""
		lvl = pygame.Surface((int(self.size.x) * Tile.size[0], int(self.size.y) * Tile.size[1]), cfg.SURFACEFLAGS)
		lvl.fill((0,0,0,255))
		
		if self.bgimage:
			lvl.blit(self.bgimage, (0,0))
		if self.image:
			lvl.blit(self.image, (0,0))
		
		
		preview = pygame.Surface((self.size.x, self.size.y), cfg.SURFACEFLAGS)
		area = pygame.Rect(0, 0, Tile.size[0], Tile.size[1])
		
		preview.lock()
		for yOffset in xrange(int(self.size.y)):
			y = yOffset * Tile.size[1]
			
			for xOffset in xrange(int(self.size.x)):
				x = xOffset * Tile.size[0]
				
				preview.set_at((xOffset, yOffset), self.floorplan[yOffset][xOffset].color)
				
				if not self.image:
					lvl.blit(self.floorplan[yOffset][xOffset].image, (x, y), area)
		
		preview.unlock()
		
		self.preview = preview
		if not self.image:
			self.image = lvl
		
		return self
	
	def save_images(self):
		"""Creates the image."""
		pygame.image.save(self.image, self.name + ".jpg")
		pygame.image.save(self.preview, self.name + " preview.png")
	
	def get_intro_image(self):
		"""Return the intro image of the level.
		
		@return the image object, None if none is supplied or it cannot be loaded."""
		return self.intro_image

class _LevelFactory(object):
	
	def __init__(self):
		self.levels = {}
		
	def get_Level(self, name):
		if not name in self.levels:
			level = Level(name);
			self.levels[name]=level
		
		return self.levels[name]

LevelFactory = _LevelFactory()

if __name__ == "__main__":
	pygame.init()
	Level("level_01").create_images().save_images()
	Level("level_02").create_images().save_images()
	Level("level_03").create_images().save_images()
	Level("level_04").create_images().save_images()
	Level("level_05").create_images().save_images()
