"""
Copyright (c) 2009 [name + "(at)student.hpi.uni-potsdam.de" for name in ["murat.knecht", "franziska.haeger", "christian.schwarz"]]

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import pygame, os

import events, cfg
from vector2D import Vector2D
from resourcemanager import ResourceManager
from tile import Tile
from weaponcfg import weaponcfgs

from drawables import Animation
from animations import WeaponAnimation, ResourceAnimation

import buffs

stone = None
earth = None

class WeaponFrameFactory(object):
	weaponFrames = {}
	
	@classmethod
	def get_animation_frames(cls, tower, level, animationImage, animationSize, isWeaponAttackAnimation=True):
		key = "%s_%s_%s" % (tower, level, isWeaponAttackAnimation)
		
		if not key in cls.weaponFrames:
			frames = []
			if not isinstance(animationImage, pygame.Surface):
				animationImage = ResourceManager.get_image(animationImage)
			
			for offset in xrange(0, animationImage.get_rect().w, animationSize[0]):
				surface = pygame.Surface(animationSize, cfg.SURFACEFLAGS)
				surface.fill((0,0,0,0))
				area = pygame.Rect(offset, 0, offset + animationSize[0], animationSize[1])
				surface.blit(animationImage, (0,0), area)
				frames.append(surface)
			
			cls.weaponFrames[key] = frames
		
		return cls.weaponFrames[key]		


	def __init__(self, center):
		if not Flower.frames:
			Flower.frames = Animation.get_animation_images(ResourceManager.get_image("menu" + os.sep + "flower.png"), 39)
		
		Animation.__init__(self, center, Flower.frames, 2000, loop=True)

class Weapon(Animation):
	"""Base class for all weaponry."""
	
	animationSize = (0,0)
	
	def __init__(self, identifier, level):
		"""Initializes the weapon.
		
		@param config Config of the weapon.
		"""
		Animation.__init__(self, loop=True, size=weaponcfgs[identifier][level]["size"])
		
		self.config = weaponcfgs[identifier]
		self.identifier = identifier
		self.level = level
		self.update_weapon(self.level)
		
		self.cooldown_left = 0 # initially, a weapon is ready to fire.
		self._show_range = False
		self.attackedPos = None
	
	def get_weapon_animation_frames(self):
		return WeaponFrameFactory.get_animation_frames(self.identifier,
														 self.level,
														 self.attAnimImg,
														 self.__class__.animationSize,
														 False)
	
	def create_buff(self, buffconfig):
		if not hasattr(buffs, buffconfig["name"]):
			return None
		
		return getattr(buffs, buffconfig["name"])(**buffconfig)

	def get_buffs(self):
		if not self.hasBuffs:
			return None
		
		buffs = []
		for buff in self.buffs:
			buffs.append(self.create_buff(buff))
		
		return buffs
	
	@property
	def hasBuffs(self):
		return self.buffs
	
	def sell_weapon(self):
		res = int(self.costs * 0.9)
		events.add_resources(res)
		events.add_animation(ResourceAnimation(self.rect.center, res))
	
	def update_weapon(self, level, resource_function=None):
		if level >= len(self.config):
			return False
		
		if resource_function:
			if not resource_function(self.config[self.level]["costs"]):
				return False
		
		self.level = level
		
		"""Updates the properties of the weapon according to the config and current upgrade level.
		
		@param upgrade_level Upgrade level of the weapon."""
		self.name       = self.config[self.level]["name"]	
		self.shortname  = self.config[self.level]["shortname"]	
		
		self.frames     = WeaponFrameFactory.get_animation_frames(self.identifier, level, self.config[level]["orgimage"], self.config[level]["size"])
		self.duration   = self.config[self.level]["animDuration"]
		self.reset_delay()
		self.frame_switch(True)
		
		self.icon       = self.config[self.level]["icon"]
		self.costs      = self.config[self.level]["costs"]
		self.cooldown   = self.config[self.level]["cooldown"]
		self.range      = self.config[self.level]["range"]
		self.damage     = self.config[self.level]["damage"]
		self.damagetype = self.config[self.level]["damagetype"]
		self.sound      = self.config[self.level]["sound"]
		self.attackables = self.config[self.level]["attackables"]
		# area of effect
		self.aoe		= self.config[self.level]["aoe"]
		if not 0 <= self.aoe <= 2:
			raise ValueError("Illegal weapon configuration: AoE must be between 0 and 2 tiles.")
		
		if "attAnimImg" in self.config[self.level]:
			self.attAnimImg = self.config[self.level]["attAnimImg"]
			self.weaponAnimationFrames = self.get_weapon_animation_frames()
		else:
			self.attAnimImg = None
			self.weaponAnimationFrames = []
		
		if "buffs" in self.config[self.level]:
			self.buffs = self.config[self.level]["buffs"]
		else:
			self.buffs = None
		
		# updates the towers image
		# if self.level > 0 :
		#	self.set_image(self.rect.center)
		
		return True
	
	@property
	def image(self):
		return self.surface

	def has_mode(self, name):
		"""Return true if the weapon has the mode.
		
		@param name The name of the mode.
		"""
		if self.modes[name]:
			return True
		
		return False
	
	def has_type(self, name):
		"""Return true if the weapon has the damage type.
		
		@param name The name of the type.
		"""
		if self.types[name]:
			return True
		
		return False
	
	def has_attackable(self, name):
		"""Return true if the weapon has the damage type.
		
		@param name The name of the type.
		"""
		if self.attackables[name]:
			return True
		
		return False
	
	def has_upgrade(self):
		"""Determine, if the weapon has a further upgraded variant."""
		return len(self.config) > self.level+1
		
	def get_upgrade_costs(self):
		"""Determine the update costs, None if no further upgrade is available."""
		if not self.has_upgrade():
			return None
		
		return self.config[self.level+1]["costs"]
	
	def update(self, tick):
		self.lastframetime += tick
		self.frame_switch()

class Tower(Weapon):
	"""The Towerbaseclass."""
	def __init__(self, towername, level=0):
		"""Initializes the tower.
		
		@param animation Sets the towers attack animation.
		"""		
		Weapon.__init__(self, towername, level)
		
		self.buildable = None
		self.hover = True		
		self.update_buildstate()
		
		# Currently targeted enemy.
		self.enemy = None
		
		# Target tile that the tower is firing on. Only considered, if the tower
		# inflicts AoE damage and the user explicitely selects the tile.
		self.target_tile = None
	
	def update(self, tick):
		# switches the animation frames
		Weapon.update(self, tick)
		# decrease cool down time
		self.cooldown_left = max(0, self.cooldown_left - tick)
	
	def build(self):
		self.surface.fill((0,0,0,0))
		self.surface.blit(self.frames[0], (0,0))
		self.hover = False
		self.update_buildstate(True)

	def update_buildstate(self, buildable=False):
		if buildable == self.buildable:
			return
		
		self.buildable = buildable
		
		if self.buildable:
			self.surface.fill((0,200,0,128))
		else:
			self.surface.fill((200,0,0,128))
		self.surface.blit(self.frames[0], (0,0))

	def sprite_in_firing_range(self, sprite):
		"""Determine if the sprite is in firing range.
		
		May be used to filter() a group of sprites.
		
		Does not consider cool down.
		The distance is calculated in terms of the euclidean distance of the center
		of the two rectangles.
		
		@param sprite a pygame.Sprite instance
		
		@return True, in case the sprite is in firing range, False otherwise
		"""
		connecting_vector = Vector2D(self.rect.center).sub(sprite.rect.center)
		distance = connecting_vector.magnitude()
		
		return distance <= self.range * Tile.size[0] - Tile.size[0] / 2
	
	def has_valid_enemy(self):
		"""Determines if the tower has an valid enemy.
		
		@return Returns True if the tower has an enemy, False otherwise.
		"""
		if not self.enemy:
			return False
		
		if not self.sprite_in_firing_range(self.enemy):
			self.enemy = None
			return False
		
		return True
	
	def draw_attack(self, surface):
		if not self.attackedPos:
			return
		
		self.add_attack_animation(surface)
		self.attackedPos = None
	
	def add_attack_animation(self, surface=None):
		raise NotImplementedError("%s has to overwrite add_attack_animation." % self.__class__.__name__)
	
	def set_enemy(self, enemy=None):
		self.enemy = enemy

class FrogTower(Tower):
	def __init__(self, level=0):
		Tower.__init__(self, "FrogTower", level)
	
	def add_attack_animation(self, surface):
		pygame.draw.line(surface, (255,0,0), self.rect.center, self.attackedPos, 5)

class FlyFlapTower(Tower):
	animationSize = (40, 40)
	
	def __init__(self, level=0):
		Tower.__init__(self, "FlyFlapTower", level)

	def add_attack_animation(self, surface=None):
		events.add_animation(WeaponAnimation(self.attackedPos,
											 self.weaponAnimationFrames,
											 1000))

class InsecticideTower(Tower):
	animationSize = (40, 40)
	
	def __init__(self, level=0):
		Tower.__init__(self, "InsecticideTower", level)

	def add_attack_animation(self, surface=None):
		events.add_animation(WeaponAnimation(self.attackedPos,
											 self.weaponAnimationFrames,
											 2000))

class HoneyPotTower(Tower):
	animationSize = (40, 40)

	def __init__(self, level=0):
		Tower.__init__(self, "HoneyPotTower", level)

	def add_attack_animation(self, surface=None):
		events.add_animation(WeaponAnimation(self.attackedPos,
											 self.weaponAnimationFrames,
											 1000))
